from qmdl.mdl import Mdl
from qmdl.palette import palette
import struct

#read the model file
mymdl = Mdl()
with open("horror_original.mdl", "rb") as infile:
    mymdl.read(infile)

# our first task is to fix all the vertex normals
# we want every vertex which occupies the same position to have the same normal
reference_frame = mymdl.frames[0]
def frame_vertex_key(i):
    return struct.pack("3B",
                       reference_frame.vertices[i].position[0],
                       reference_frame.vertices[i].position[1],
                       reference_frame.vertices[i].position[2])
# We will use the above function as the key to this dictionary
distinct_frame_vertices = {}
# In the dictionary will be the indices of all the vertices which are in the
# same position in the reference pose

# First set up an empty list for each distinct vertex
# Multiple vertices may map to a distinct vertex
# but we don't mind yet because it just overwrites an empty list
for i in range(len(mymdl.vertices)):
    distinct_frame_vertices[frame_vertex_key(i)] = []
# add the vertices to each list
for i in range(len(mymdl.vertices)):
    distinct_frame_vertices[frame_vertex_key(i)].append(i)
for frame in mymdl.frames:
    # for each frame we visit the sets of intersecting vertices
    for coincident_vertices in distinct_frame_vertices.values():
        #make a list of all the normals
        normals = []
        for i in coincident_vertices:
            normals.append(frame.vertices[i].decode())
        # sum all the normals
        average_normal = tuple(sum(t) for t in zip(*normals))                                                                       
        # we don't have to normalize it to encode it
        for i in coincident_vertices:
            frame.vertices[i].encode(average_normal)

# Now vertices are considered the same if:
# a) they have the same skin coordinates
# b) they have the same coordinates in the reference frame
# If we tried to merge them without condition a) it would ruin the skinmap

def vertex_key(i):
    return struct.pack("2H3B",
                       mymdl.vertices[i].u,
                       mymdl.vertices[i].v,
                       reference_frame.vertices[i].position[0],
                       reference_frame.vertices[i].position[1],
                       reference_frame.vertices[i].position[2])
distinct_vertices = {}

for i in range(len(mymdl.vertices)):
    distinct_vertices[vertex_key(i)] = []
    
for i in range(len(mymdl.vertices)):
    distinct_vertices[vertex_key(i)].append(i)

# Now each distinct_vertices entry has a list of vertices which will map to it
# Invert it, so we have a list of the new vertex entry for each of the old ones
distinct_keys = list(distinct_vertices.keys())
# create an empty list the length of the vertex list
vertex_mapping = [None for i in range(len(mymdl.vertices))]
for i in range(len(distinct_keys)):
    for vertex in distinct_vertices[distinct_keys[i]]:
        vertex_mapping[vertex] = i
        
# We now need to collapse all of the mergable vertices into one
# This happens in three places:
# In the vertex list, we output one u-v coordinate for each unique vertex
# In the triangle list, we replace the indices according to vertex_mapping
# In the frames, we output one vertex from each mergable set. Because we have
# unified the normals earlier, we can just take the first vertex as
# representative

new_vertices = []
for vertex_list in distinct_vertices.values():
    new_vertices.append(mymdl.vertices[vertex_list[0]])
mymdl.vertices = new_vertices

for triangle in mymdl.triangles:
    triangle.vertices = tuple(vertex_mapping[i] for i in triangle.vertices)

for frame in mymdl.frames:
    new_vertices = []
    for vertex_list in distinct_vertices.values():
        new_vertices.append(frame.vertices[vertex_list[0]])
    frame.vertices = new_vertices

#write the model file
with open("horror.mdl", "wb") as outfile:
    mymdl.write(outfile)
